# Federated Learning Network - P2P version

Federated Learning network - P2P version is the evolution of the Msc capstone project from Eduardo Yáñez Parareda.
In this version every node is considered equals to other, so there's no central server and clients, just nodes.
A Federated Learning Network is the one where Machine Learning compute is happening inside the nodes and there's no data transfer, just
ML model weight transfers between nodes. In this way, privacy regarding the used data is ensured. The average to the result of all received model weights
from the rest of the nodes, which is formally named Federated Average. That average will be used by the nodes on the next rounds of trainings.

## Introduction

There are two options for running the nodes in the network, using Docker to create containers with a node each, or
using a standard local installation from the command line. Of course, both can be mixed (some nodes running as containers and other by command line).

### Datasets

For now, there are three models for training: MNIST, Chest X-Ray and CIFAR-10. For using the MNIST one
and CIFAR-10 you don't need to install anything else because the node downloads the datasets when it runs the training, 
but for the Chest X-Ray model you'll need a dataset to get it working.

Actually, if you're planning to run several nodes without Docker, it's best to run the unit tests of the project at least
once so the datasets are downloaded before you start spawning more than one node. Errors will arise due to having several nodes writing to the same dataset directory when they try to download the same datasets. If datasets are pre-downloaded before training, no issue is expected to happen in that regard. Please, check how to run unit tests below in this README.

Download the dataset from https://data.mendeley.com/public-files/datasets/rscbjbr9sj/files/f12eaf6d-6023-432f-acc9-80c9d7393433/file_downloaded, 
and uncompress it wherever you want on all node machines, in a folder called `chest_xray`. 
The final structure must be (other content in this folder will be ignored):

    chest_xray
         |----- train
         |        |----- NORMAL
         |        |----- PNEUMONIA
         |
         |----- test
                  |----- NORMAL
                  |----- PNEUMONIA

By default, the node looks for it at DATASETS_PATH/chest_xray. The variable DATASET_PATH is defined 
in the configuration file `src/configurations/config.py`.

If you're going to run the node using Docker, you must pass a volume as a container parameter to indicate where you 
have the datasets:

    -v /your_dataset_folder:/datasets

In particular, for Chest X-Ray training, it'll expect a directory `chest_xray` in your dataset's directory with at least 
two folders `train` and `test` with x-ray images.

Example: You downloaded the Chest X-Ray training dataset into /home/johndoe/datasets/chest_xray, for Docker you need to use
this argument:

    -v /home/johndoe/datasets:/datasets

### Docker

Create the Docker image of the server by doing this:
    
    docker build -t fl-node-image -f Dockerfile .
   
#### Running the project   

**IMPORTANT**: To be able to use the Chest X-Ray model training follow the instructions of _Training the Chest X-Ray model_ section.

Here's an example for running several nodes in your machine using Docker. First, assuming your port 5000 in your machine is free to be opened by any app and the IP of your network interface is 192.168.1.45:

    docker run --rm -e 'FLNET_NODE_PORT=5000' -p 5000:5000 -e 'FLNET_NODE_IP=192.168.1.45' -v /your_dataset_folder:/datasets fl-node-image:latest

Make sure to change /your_dataset_folder with the path to your downloaded datasets (check "datasets" above, there's an example at the end).

To run another node in the same machine, just do this:
    
    docker run --rm -e 'FLNET_NODE_PORT=5001' -p 5001:5001 -e 'FLNET_NODE_IP=192.168.1.45' -e 'FLNET_BOOTSTRAP_PAIRS=192.168.1.45:5000' -v /your_dataset_folder:/datasets fl-node-image:latest

This will open a node at port 5001 and will try to connect to your previous node at 192.168.1.45:5000.

You can open more nodes if you want to:
    
    docker run --rm -e 'FLNET_NODE_PORT=5002' -p 5002:5002 -e 'FLNET_NODE_IP=192.168.1.45' -e 'FLNET_BOOTSTRAP_PAIRS=192.168.1.45:5000' -v /your_dataset_folder:/datasets fl-node-image:latest
    docker run --rm -e 'FLNET_NODE_PORT=5003' -p 5003:5003 -e 'FLNET_NODE_IP=192.168.1.45' -e 'FLNET_BOOTSTRAP_PAIRS=192.168.1.45:5000' -v /your_dataset_folder:/datasets fl-node-image:latest
    docker run --rm -e 'FLNET_NODE_PORT=5004' -p 5004:5004 -e 'FLNET_NODE_IP=192.168.1.45' -e 'FLNET_BOOTSTRAP_PAIRS=192.168.1.45:5000' -v /your_dataset_folder:/datasets fl-node-image:latest

Just make sure always that FLNET_NODE_PORT is different for every spawned node or your new node will try to occupy an already occupied port in your machine.
FLNET_NODE_IP environment variable is optional under other setups / network requirements. Without it, what the app will do is that when a request comes to a node, it will get its public IP directly from the request. FLNET_NODE_IP will just force announcing to the rest of the nodes a particular IP address.

FLNET_NODE_IP is specially neccesary when using Docker on Windows or Mac due to how they work for making a namespace / isolated / containerized network for containers.

##### Note for Linux users

Always use the "host" network mode from Docker. That way you'll avoid odd networking issues. Nodes gather the IP address from the requests made to them 
to determine the public IP of the nodes, so with the "host" network mode, there won't be any issue with this mechanism since the node will use the same resources as
your host machine. It will be share the same privileges / resources without any virtualization / containerazation as your other apps. No need to use FLNET_NODE_IP in this mode to spawn nodes in the same machine:

    docker run --network='host' --rm -e 'FLNET_NODE_PORT=5002' -e 'FLNET_BOOTSTRAP_PAIRS=127.0.0.1:5000' -v /your_dataset_folder:/datasets fl-node-image:latest

### Bare bones

If Docker is not an option, then you must install everything and running from the command line.
Python version must be the latest minor version of the 3.8.x series.

One way is using miniconda or conda, just do inside the src folder:

    conda create --name flnet python=3.8
    conda activate flnet

Once you're ready to install packages, do this inside the src folder too:

    pip install -r requirements.txt

As alternative, instead of `conda activate` you can just use pipenv if you have it installed (like before, inside the src folder):

    pipenv install

That will make an isolated environment and will install all the requirements at the same time.

Just make sure, that pipenv is using python 3.8. If you have several python version installed in your system, you need to check its documentation to tell it to use python 3.8.

#### Running the project   

That's very simple, just go to `src` folder and execute (Linux / Mac):

    FLNET_NODE_PORT=8080 python flnet.py

On Windows, environment variables depend on the terminal you use and may need to be setup before running flnet.py, e.g.: 

* In CMD:

        set ENVVAR=value

* In PowerShell:

        $env:ENVVAR=value

If you use pipenv, do before above's command line, the following command to enter in the shell prepared for you by pipenv:

    pipenv shell

Once flnet.py is running, it'll start an isolated node in `http://localhost:8080`. To see that's running well, open a browser and go to that URL.
You'll see the dashboard of the network.

To bootstrap the next node you open, run it like this:

    FLNET_BOOTSTRAP_PAIRS='127.0.0.1:8080' FLNET_NODE_PORT=8081 python flnet.py

That means that the new node will try to connect to 127.0.0.1:8080 at startup and will open the 8081 port. Once that happens, both nodes will know each other and will share all of their known nodes to each other (which at this point, it's none. This node sharing is effective when more than two nodes are instantiated).

Of course, you can bootstrap to more than one node at startup if you want to by separating them with commas: FLNET_BOOTSTRAP_PAIRS='127.0.0.1:8080,127.0.0.1:6000'.
That will try to connect to one node at port 8080 in your machine and another node at your machine at port 6000.

There's another environment variable called FLNET_NODE_IP you may need. Read the Docker part above to know more about that one.

Once they're connected, click on the refresh icon or press F5 in your keyboard in the dashboard and you'll see all the known nodes.

## Training sessions

Once we have several nodes connected betweent them, just open the server dashboard and click on 
the _Do a training round_ button.

This action will launch a training session between all the nodes registered. You can see the progress of the training in each 
node's console. For example, for MNIST training you will see something like this in the node console:

```
[2021-05-02 20:07:39,816] {p2p_controller.py:28} INFO - start_training_in_this_node() method called...
[2021-05-02 20:07:39,816] {training_service.py:27} INFO - Training requested. Started...
[2021-05-02 20:07:39,816] {node_service.py:87} INFO - Node status changed to: TRAINING
[2021-05-02 20:07:39,817] {mnist_trainer.py:116} INFO - Loading dataset MNIST_SAMPLE...
[2021-05-02 20:07:40,110] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 1: 0.56
[2021-05-02 20:07:40,111] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 2: 0.56
[2021-05-02 20:07:40,112] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 3: 0.6
[2021-05-02 20:07:40,114] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 4: 0.6
[2021-05-02 20:07:40,115] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 5: 0.6
[2021-05-02 20:07:40,116] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 6: 0.68
[2021-05-02 20:07:40,117] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 7: 0.68
[2021-05-02 20:07:40,118] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 8: 0.68
[2021-05-02 20:07:40,120] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 9: 0.68
[2021-05-02 20:07:40,121] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 10: 0.68
[2021-05-02 20:07:40,122] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 11: 0.72
[2021-05-02 20:07:40,123] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 12: 0.84
[2021-05-02 20:07:40,124] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 13: 0.88
[2021-05-02 20:07:40,125] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 14: 0.92
[2021-05-02 20:07:40,127] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 15: 0.96
[2021-05-02 20:07:40,128] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 16: 0.96
[2021-05-02 20:07:40,129] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 17: 0.96
[2021-05-02 20:07:40,130] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 18: 0.96
[2021-05-02 20:07:40,131] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 19: 0.96
[2021-05-02 20:07:40,132] {mnist_trainer.py:45} INFO - Accuracy of model trained at epoch 20: 0.92
[2021-05-02 20:07:40,133] {training_service.py:35} INFO - Last training requested for MNIST finished...
[2021-05-02 20:07:40,133] {node_service.py:87} INFO - Node status changed to: SENDING_MODEL_PARAMS_TO_NODES
[2021-05-02 20:07:40,133] {training_service.py:38} INFO - Broadcasting new model params of MNIST to nodes...
[2021-05-02 20:07:40,133] {node_service.py:148} INFO - Doing POST call to http://127.0.0.1:4000/training/m
```

You can do more training sessions afterwards and see how the model improves.

## Customization

You can see and change some training parameters (epochs, batch size and learning rate) per dataset trainer at:

        src/trainers/<trainer>.py get_default_federated_learning_params method

## Unit tests

 There are serveral unit tests done and they can be executed with this command line inside the src folder:
    python -m unittest

They'll test some aspects of the code and they'll do some training, validating that they'll work on your machine.

## Known issues

* The P2P layer is not existent as such. It's just a simulation done through a REST API layer. There's quite a lot of improvements to be done there.
* Interestingly enough, the order of python module imports is quite important. In `src/trainers/__init__.py`, if you change the order of the imports, it's likely
that you'll trigger a segmentation fault. It's very likely that all these Machine Learning Python libs aren't supposed to be loaded at the same time since they're usually in use isolated. For a practical application of this software, it's recommended therefore to unregister the datasets you aren't going to use by removing the unneeded trainers from the registered_trainers in `src/trainers/__init__.py`, removing the corresponding import and removing the unneeed trainer python files themselves.
* Please, check the tmp folder from time to time given that when the app doesn't exit gracefully, it will leave files and dirs in there. There's no tmp folder cleaning at startup too. Automatic dataset loading for some trainers install datasets in other folders of your system.
* There's no persistence regarding the trained models, therefore their calculations inside each node will be gone once the node ceases to exist.
