import logging
import sys
import os
import uuid
import asyncio
from typing import Dict
from concurrent.futures import ThreadPoolExecutor

# Default config you may change as you wish
LOGGING_LEVEL = logging.INFO
TMP_PATH = '../tmp'
DATASETS_PATH = '../datasets'
REQUEST_TIMEOUT_IN_SECONDS = 30
BOOTSTRAP_WAIT_IN_SECONDS = 2
PRUNE_UNRESPONSIVE_NODES = False
WAIT_TIME_IN_SECONDS_TO_FORCE_FEDAVG = 30
WAIT_TIME_IN_SECONDS_TO_FORCE_FEDAVG_MARGIN = 20
RANDOM_SEED = 689532449


# These variables will change in run-time following environment variables. Don't touch them.
BOOTSTRAP_PAIRS = None
NODE_ID = None
NODE_IP = None
NODE_PORT = None

APP_THREAD_POOL_EXECUTOR = ThreadPoolExecutor(max_workers=1)
FORCE_FEDAVG_TIMERS: Dict[str, asyncio.Task] = {}

def logging_config():
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.addFilter(lambda record: record.levelno < logging.ERROR)
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(logging.ERROR)
    handlers = [stdout_handler, stderr_handler]

    logging.basicConfig(
        level=LOGGING_LEVEL,
        format='[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s',
        handlers=handlers
    )


def extract_vars_from_env():
    global BOOTSTRAP_PAIRS
    global NODE_PORT
    global NODE_IP
    global DATASETS_PATH
    global NODE_ID

    BOOTSTRAP_PAIRS = os.environ.get('FLNET_BOOTSTRAP_PAIRS')
    if BOOTSTRAP_PAIRS is not None:
        BOOTSTRAP_PAIRS = BOOTSTRAP_PAIRS.split(',')

    NODE_PORT = os.environ.get('FLNET_NODE_PORT')
    if NODE_PORT is None:
        print('\nFLNET_NODE_PORT env var missing! With this environment variable, you indicate which port is going '
              'to be opened.\n')
        sys.exit(1)

    if NODE_PORT == 0:
        print('\nFLNET_NODE_PORT cannot be 0! We cannot open a random port by design.\n')
        sys.exit(1)

    NODE_IP = os.environ.get('FLNET_NODE_IP')

    datasets_path = os.environ.get('FLNET_DATASETS_PATH')
    if datasets_path is not None:
        DATASETS_PATH = datasets_path

    NODE_ID = os.environ.get('FLNET_NODE_ID')
    
    
def generate_node_id():
    global NODE_ID

    if NODE_ID is None:
        NODE_ID = uuid.uuid1()


def app_init_configuration():
    logging_config()
    extract_vars_from_env()
    generate_node_id()
