import asyncio
import configuration.config as config
import services.job_service as job_service
import services.node_service as node_service
from quart import Quart
from hypercorn.config import Config as ServerConfig
from hypercorn.asyncio import serve


async def prepare_server_and_bootstrapping():
    app = Quart(__name__, static_url_path="", static_folder="resources/static")

    from controllers.ui_controller import ui_controller
    from controllers.p2p_controller import p2p_controller
    app.register_blueprint(ui_controller)
    app.register_blueprint(p2p_controller)

    server_config = ServerConfig()
    server_config.bind = ['0.0.0.0:' + str(config.NODE_PORT)]

    server = asyncio.create_task(serve(app, server_config))
    asyncio.create_task(node_service.bootstrap_to())
    await server


def startup():
    config.app_init_configuration()

    print("\nFederated Learning Network. P2P version.\n")
    print("© Eduardo Yáñez Parareda, Tomás García Pérez. Universitat Oberta de Catalunya.")
    print("------------------------------------------------------------------------------\n")
    print("Configuration")
    print("-------------")
    print("Node Id: {}".format(config.NODE_ID))
    print('Bootstrap ip/port pairs: {}'.format(
        config.BOOTSTRAP_PAIRS if config.BOOTSTRAP_PAIRS is not None else 'Not defined. Initial node.'))
    print('Node port: {}'.format(
        config.NODE_PORT if config.NODE_PORT != 0 else 'Not defined. Random port will be selected.'))
    print('Node ip: {}'.format(config.NODE_IP if config.NODE_IP is not None else 'Public IP remote address of requests in use'))
    print('Datasets directory: {}'.format(config.DATASETS_PATH))
    print("\nInitializing app...\n")

    job_service.prepare_jobs_at_startup()

    asyncio.run(prepare_server_and_bootstrapping())

    job_service.clean_jobs()


if __name__ == '__main__':
    startup()