import logging

from typing import List

from trainers import register_trainers
from models.job_model import JobModel

job_service_logger = logging.getLogger()

__JOBS : List[JobModel] = []


def prepare_jobs_at_startup():
    registered_trainers = register_trainers()
    for trainer in registered_trainers:
        job_service_logger.info('Registering {} ({}) Trainer...'.format(trainer.get_label_name(), trainer.get_id()))
        __JOBS.append(JobModel(trainer=trainer))


def obtain_available_trainers_id_labels():
    global __JOBS

    result = []
    for job in __JOBS:
        result.append({'id': job.get_trainer().get_id(), 'label': job.get_trainer().get_label_name()})
    return result


def find_job_by_trainer_id(trainer_id):
    global __JOBS

    job = [job for job in __JOBS if job.get_trainer().get_id() == trainer_id]
    if len(job) == 0 or len(job) > 1:
        raise ValueError('Not expected result from job search: ' + str(len(job)))

    return job[0]


def obtain_current_info_per_trainer_for_ui():
    global __JOBS

    result = []
    for job in __JOBS:
        result.append(job.to_info_for_ui())

    return result


def clean_jobs():
    global __JOBS

    for job in __JOBS:
        job.get_trainer().cleanup()
