import asyncio
import logging
import services.node_service as node_service
import services.job_service as job_service
from services.training_service import TrainingService
from configuration.constants.http_verb import HttpVerb
from models.job_model import JobModel


class UserInterfaceService:
    def __init__(self):
        self.logger = logging.getLogger()
        self.training_service = TrainingService()

    async def tell_network_to_train_and_start_training_in_this_node_too(self, job: JobModel):
        self.logger.info('There are currently {} nodes'.format(node_service.obtain_amount_of_known_nodes()))

        self.logger.info('Requesting to the rest of nodes to train model: {}'.format(job.get_trainer().get_id()))

        job.increase_round()

        await node_service.broadcast_message(method=HttpVerb.POST, url_path='/training',
                                             json_body=job.to_current_job_state_message_body())

        asyncio.create_task(self.training_service.start_training_in_this_node(job))

    def obtain_info(self):
        return {'iAm': node_service.get_this_node().to_request_body(),
                'lastNodeStatus': node_service.get_status(),
                'jobsInfo': job_service.obtain_current_info_per_trainer_for_ui()}
