import unittest
import numpy
import json
from trainers.chest_x_ray_pneumonia_trainer import ChestXRayPneumoniaTrainer


class ChestXRayPneumoniaTrainerTest(unittest.TestCase):

    def test_trainer(self):
        trainer = ChestXRayPneumoniaTrainer()
        trainer.prepare_initial_model_params()
        trainer.train_model(trainer.get_default_federated_learning_params())
        trainer.update_trainer_model_accuracy_pre_federated_avg(trainer.get_default_federated_learning_params())
        print('MODEL SHAPE: {}'.format(numpy.array(trainer.model_params).shape))
        print('ACCURACY: {}'.format(trainer.get_pre_fedavg_accuracy()))
        print('RE-CALCULATED ACCURACY: {}'.format(trainer.calculate_model_accuracy(
            trainer.get_default_federated_learning_params(), trainer.model_params)))

    def test_model_params_transfer(self):
        trainer = ChestXRayPneumoniaTrainer()
        trainer.prepare_initial_model_params()
        to_request_body_before = trainer.generate_weight_params_message_body()
        trainer.set_model_params(trainer.process_weight_params_message_body(to_request_body_before))
        to_request_body_after = trainer.generate_weight_params_message_body()
        # Comparison in String just to be 100% sure
        assert json.dumps(to_request_body_before['weights']) == json.dumps(to_request_body_after['weights'])

    def test_fedavg(self):
        trainer = ChestXRayPneumoniaTrainer()
        trainer.set_model_params(numpy.ones((6, 3, 3, 3, 32), dtype=numpy.float32))
        to_request_body_model = trainer.generate_weight_params_message_body()
        trainer.set_model_params(numpy.zeros((6, 3, 3, 3, 32), dtype=numpy.float32))
        trainer.update_model_params_doing_federated_average([to_request_body_model['weights'],
                                                             trainer.generate_weight_params_message_body()['weights']])
        model_params = trainer.generate_weight_params_message_body()
        assert model_params['weights'][0][0][0][0][0] == 0.3333333333333333
        assert model_params['weights'][5][1][2][2][25] == 0.3333333333333333


if __name__ == '__main__':
    unittest.main()
