import unittest
import numpy
import json
import torch
from trainers.mnist_trainer import MnistTrainer


def train_and_calculate_accuracy(trainer):
    trainer.train_model(trainer.get_default_federated_learning_params())
    trainer.update_trainer_model_accuracy_pre_federated_avg(trainer.get_default_federated_learning_params())
    print('MODEL SHAPE: {}'.format(numpy.array(trainer.model_params).shape))
    print('ACCURACY: {}'.format(trainer.get_pre_fedavg_accuracy()))
    print('RE-CALCULATED ACCURACY: {}'.format(trainer.calculate_model_accuracy(
        trainer.get_default_federated_learning_params(), trainer.model_params)))

class MnistTrainerTest(unittest.TestCase):

    def test_trainer(self):
        trainer = MnistTrainer()
        trainer.prepare_initial_model_params()
        for i in range(20):
            train_and_calculate_accuracy(trainer)

    def test_model_params_transfer(self):
        trainer = MnistTrainer()
        trainer.prepare_initial_model_params()
        to_request_body_before = trainer.generate_weight_params_message_body()
        trainer.set_model_params(trainer.process_weight_params_message_body(to_request_body_before))
        to_request_body_after = trainer.generate_weight_params_message_body()
        # Comparison in String just to be 100% sure
        assert json.dumps(to_request_body_before['weights']) == json.dumps(to_request_body_after['weights'])
        assert json.dumps(to_request_body_before['bias']) == json.dumps(to_request_body_after['bias'])

    def test_fedavg(self):
        trainer = MnistTrainer()
        trainer.set_model_params((torch.tensor([1, 2], dtype=torch.float, requires_grad=True),
                                  torch.tensor([2, 3], dtype=torch.float, requires_grad=True)))
        trainer.update_model_params_doing_federated_average([
            (torch.tensor([2, 3], dtype=torch.float, requires_grad=True),
                                  torch.tensor([3, 4], dtype=torch.float, requires_grad=True))
        ])
        model_params = trainer.generate_weight_params_message_body()
        assert model_params['weights'][0] == 1.5
        assert model_params['weights'][1] == 2.5
        assert model_params['bias'][0] == 2.5
        assert model_params['bias'][1] == 3.5


if __name__ == '__main__':
    unittest.main()
