import unittest
import services.node_service as node_service
from models.node_model import NodeModel


def _test_node_ops():
    node_service.clear_nodes()

    node = NodeModel(node_id='1.1', port='2323')

    new_nodes, unknown_nodes_to_sender = node_service.__obtain_node_differences([node])

    assert new_nodes == {node}, 'Node to be added is returned'

    node_service.add_nodes(node)
    assert node_service.__NODES == {node}, 'Current node list isnt the first node'

    new_nodes, unknown_nodes_to_sender = node_service.__obtain_node_differences([node])
    assert new_nodes == set(), 'No new node must appear here since it was a repeated insertion'
    assert unknown_nodes_to_sender == set(), 'No unknown nodes to sender since it was a repeated insertion'

    node_service.add_nodes(node)
    assert node_service.__NODES == {node}, 'Current node list should be just the first node'

    node_service.remove_node(node)
    assert node_service.__NODES == set(), 'Node list is not empty: {}'.format(node_service.__NODES)

    node_service.add_nodes(node)

    node2 = NodeModel(node_id='1.2', port='3232')
    new_nodes, unknown_nodes_to_sender = node_service.__obtain_node_differences([node2])
    assert new_nodes == {node2}, 'Node to be added is returned'
    assert unknown_nodes_to_sender == {node}, 'Sender sent just node2, so node1 should be unknown to it'

    node_service.add_nodes(node2)
    assert node_service.__NODES == {node, node2}, 'Current node list isnt the first and second node: {}' \
        .format(node_service.__NODES)

    node_service.clear_nodes()
    node1 = NodeModel(node_id='1.1', ip='1', port='3232')
    node2 = NodeModel(node_id='1.2', ip='1', port='3232')
    node_service.add_nodes(node1)
    new_nodes, unknown_nodes_to_sender = node_service.__obtain_node_differences([node1, node2])
    assert new_nodes == set(), 'No new nodes should be returned'
    assert unknown_nodes_to_sender == set(), 'Sender sent node 1 and node 2, so no node should be unknown'
    node_service.add_nodes(node2)
    assert len(node_service.__NODES) == 1, 'Node amount should still be one'


def _test_node_hash():
    node_service.clear_nodes()

    node1 = NodeModel(node_id='1.1', ip='127.0.0.1', port='2323')
    node_service.add_nodes(node1)

    node2 = NodeModel(node_id='1.2', ip='127.0.0.1', port='2323')
    node_service.add_nodes(node2)

    assert len(node_service.__NODES) == 1, 'Two nodes with same port / ip but different id were added and it shouldnt'


def _test_request_body():
    node_service.clear_nodes()

    node1 = NodeModel(node_id='1.1', ip='8768675', port='2323')
    node2 = NodeModel(node_id='1.3', ip='6543344', port='45421')

    node_service.add_nodes(node1)
    node_service.add_nodes(node2)

    result = node_service.__generate_announce_request_body()
    assert result is not None, 'There should be a result'
    assert result['iAm'] is not None, 'It should contain an iAm key'
    assert result['myKnownPeers'] is not None and len(
        result['myKnownPeers']) == 2, 'It should contain a myKnownPeers key with two elements inside'


class NodeTest(unittest.TestCase):
    def test_node_ops(self):
        _test_node_ops()

    def test_node_hash(self):
        _test_node_hash()

    def test_request_body(self):
        _test_request_body()


if __name__ == '__main__':
    unittest.main()
